/*
** Module   : LoginClient.java
** Abstract : implements client side MAJIC custom authentication logic
**
** Copyright (c) 2005-2011, TIMCO Aviation Services, Inc.
** ALL RIGHTS RESERVED.
**
**               TIMCO Aviation Services, Inc.
**                      CONFIDENTIAL
**
** -#- -I- --Date-- --JPRM-- ----------------Description-----------------
** 001 NVS 20050901   @22685 Created the initial version. This class is
**                           a custom MAJIC authenticator. It provides two
**                           sides: client side login screen, based on
**                           CHUI, and the server side authentication hook
**                           that handles the input.
** 002 NVS 20050913   @22693 Implemented retriable authentication. This
**                           class keeps the windowing environment between
**                           retries. The server controls the number of
**                           available retries.
** 003 NVS 20050914   @22712 Made the code compliant with the changes to
**                           the Authenticator interface. Removed call to
**                           suspendCharva() and added new methods instead
**                           disableInput() and enableInput(), which work
**                           at widgets level.
** 004 NVS 20050919   @22753 Completed the clientFinalize() method, that
**                           does the due cleanup now.
** 005 NVS 20051003   @22925 Prevents class loading if no CHUI was init'd
**                           and provides a readable diagnostic output.
** 006 NVS 20051220   @23771 Rewritten to match the single-threaded nature
**                           of the new CHARVA architecture. Acts as a
**                           micro server and drives the logon screen i/o.
** 007 NVS 20060114   @23911 Got rid of setScreenValue() calls. Using a
**                           ScreenBuffer in view() call instead.
** 008 NVS 20060114   @24148 waitFor signature changed in ClientExports.
** 009 GES 20060306   @24911 Name change.
** 010 NVS 20060307   @24920 Login processing should catch all condition
**                           exceptions that now can be thrown during
**                           waitFor() processing.
** 011 NVS 20060308   @24943 Using new ScreenBuffer which keeps state.
** 012 NVS 20060314   @25064 ScreenBuffer needs to have exact number of
**                           widgets as its size.
** 013 NVS 20060324   @25237 Fixed a regression that caused the QUIT uid
**                           to be handled as a regular ID.
** 014 NVS 20060417   @25560 Fixed extra blanks appearing in the entered
**                           password.
** 015 GES 20060509   @26050 Rewrote the code to use 2 frames and a PUT
**                           SCREEN like the original.  This was needed
**                           because the 1 frame approach had hard-coded
**                           positions that only worked for very wide
**                           screens.  In the common case of an 80 column
**                           screen width, the result was unusable. Many
**                           other coding convention and minor text fixes
**                           were also made.
** 016 GES 20060531   @26843 Match signature change in client exports.
** 017 GES 20060614   @27192 Match another signature change in client
**                           exports.
** 018 GES 20060619   @27455 We should not pass an invalid screen buffer
**                           to waitFor. Required to match changes in
**                           the thin client.
** 019 GES 20060627   @27607 Match another signature change in client
**                           exports enable().
** 020 EVL 20060627   @27970 Fix for misplaced logo static text for
**                           invalid login attempt. The row value must be
**                           declared explicitly.
** 021 NVS 20060713   @28030 Added painting control brackets.
** 022 GES 20060808   @28483 Match signature change in ClientExports.
** 023 GES 20061003   @30102 Added return code processing for retry case
**                           and misc cleanups.
** 024 GES 20061127   @31413 Match signature change in ClientExports.
** 025 GES 20061220   @31808 Match Authenticator interface change.
** 026 NVS 20070118   @31895 EventList class is used instead of WaitList.
** 027 SIY 20070119   @31903 Match signature change in ClientExports.
** 028 NVS 20070724   @34690 Yet another fix in ClientExports needed code
**                           change here.
** 029 NVS 20080320   @37623 Changed call to enable() export of ThinClient
**                           to match the new signature.
** 030 CA  20080919   @39859 Modified login screen to match new copyright
**                           notices.
** 031 SVL 20090424   @41925 Insert indicator is refreshed on relogin.
** 032 SIY 20090610   @42638 Notify ThinClient about current working mode.
** 033 NVS 20090816   @43643 Quitting the login panel notifies the server this
**                           is intentional exit. The plugin should return
**                           an empty array of authentication data in this case.
** 034 SIY 20100619          Switching to own implementation of character UI.
** 035 SIY 20100804          Changes in imports, some refactorings.
** 036 SVL 20100923          Changed to conform the new Authenticator API.
** 037 SIY 20110109          Changes related to new color handling code.
** 038 OM  20131118          Implemented the new method of the Authenticator interface.
*/

package aero.timco.majic.security;

import com.goldencode.p2j.net.SessionListener;
import com.goldencode.p2j.security.*;
import com.goldencode.p2j.security.SecurityManager;
import com.goldencode.p2j.ui.*;
import com.goldencode.p2j.ui.chui.*;
import com.goldencode.p2j.ui.client.*;
import com.goldencode.p2j.ui.client.event.*;
import com.goldencode.p2j.util.*;
import java.util.*;

/**
 * The class to show the TIMCO splash screen and to request the user's
 * authentication. The userID and the password should be entered into the
 * entry dialog box fields.
 */
public class LoginClient
implements Authenticator
{
   /** Client interface. */
   private ThinClient client = null; 

   /** Logo frame definition. */
   private ScreenDefinition logosd = null; 

   /** Edit frame definition. */
   private ScreenDefinition editsd = null; 
   
   /** Logo frame ID. */
   private static final int LOGO_FID = 100; 

   /** Edit frame ID. */
   private static final int EDIT_FID = 200; 
   
   /** ID of the first logo screen text. */
   private static final int LOGO_ID = LOGO_FID + 1; 

   /** User ID field ID. */
   private static final int USER_ID = EDIT_FID + 1;

   /** Password field ID. */                            
   private static final int PASSWORD_ID = USER_ID + 2; // 1 + skip

   /** Info text ID. */
   private static final int INFO_ID = PASSWORD_ID + 2; // 1 + skip

   /** Standard login status message. */
   private static final String STD_MSG = 
      "Please Login to the MAJIC System.             ";

   /** Invalid password status message. */
   private static final String INVALID_PASSWORD = 
      "                                Invalid Login!";
   
   /** Invalid user status message. */
   private static final String INVALID_USER = 
      "                User NOT Authorized to enter the MAJIC System!";

   /** splash screen text */
   private static final String[] hdr = 
   {
      "     MMMMMM           MMMMMM         ##############                           ",
      "     MMMMMMM         MMMMMMM             ###    ## ###    ### ####   ####     ",
      "     MMMMMMMM       MMMMMMMM            ###    ## ####  #### ##  ## ##  #     ",
      "     MMMMMMMMM     MMMMMMMMM           ###    ## ########## ##     ##  ##     ",
      "     MMMMMMMMMM   MMMMMMMMMM          ###    ## ### ## ###  ##  ## #  ##      ",
      "     MMMMMMMMMMM MMMMMMMMMMM         ###    ## ###    ###   #####  ####       ",
      "     MMMMMMMMMMMMMMMMMMMMMMM               Aviation Services, Inc.            ",
      "     MMMMMM MMMMMMMMM MMMMMM                                                  ",
      "     MMMMMM  MMMMMMM  MMMMMM       A          JJJJJJJJ IIIIIIII   CCCCCCC     ",
      "     MMMMMM   MMMMM   MMMMMM      AAA           JJJJ     IIII    CCCC CCCC    ",
      "     MMMMMM    MMM    MMMMMM     AAAAA          JJJJ     IIII   CCCC   CCCC   ",
      "     MMMMMM     M     MMMMMM    AAA AAA         JJJJ     IIII   CCCC          ",
      "     MMMMMM           MMMMMM   AAA   AAA  JJJJ  JJJJ     IIII   CCCC   CCCC   ",
      "     MMMMMM           MMMMMM  AAAAAAAAAAA  JJJJ JJJJ     IIII    CCCC CCCC    ",
      "     MMMMMM           MMMMMM AAAAA   AAAAA  JJJJJJJ    IIIIIIII   CCCCCCC     ",
      "     Copyright 1991 - 2007 by TIMCO Aviation Services, Inc.                   ",
      "                                                                              ",
      "                                                                              "
   };

   /** The screen buffer for the logo frame. */
   private ScreenBuffer logosb = new ScreenBuffer(LOGO_FID, hdr.length);
   
   /** The screen buffer for the editing frame. */
   private ScreenBuffer editsb = new ScreenBuffer(EDIT_FID, 3);
   
   /** The userid typed last time this was used. */
   private String lastid = null;
      
   /** The password typed last time this was used. */
   private String lastpw = null;

   /**
    * Default constructor. 
    */
   public LoginClient()
   {
      // locate the ThinClient
      client = ThinClient.getInstance();
   }

   /**
    * Implements client side custom authentication logic.
    *
    * @param    parameters
    *           Additional configuration parameters. This plugin uses
    *           "option" parameter which is taken from the directory.
    * @param    code
    *           The result of the most recent attempt to authenticate or
    *           <code>AUTH_RESULT_NONE</code> if this is the first attempt.
    *
    * @return   Array of bytes to be transmitted to the server as the
    *           authorization input. 
    */
   @Override
   public byte[] clientAuthHook(Map<String, Object> parameters, int code)
   {
      OutputManager tk = OutputManager.instance();
      String option = "";
      if (parameters != null)
         option = (String) parameters.get("option");
      
      if (option == null)
         option = "";
      
      // push the frame definition the first time in
      if (code == AUTH_RESULT_NONE)
      {
         int maxlen = 0;
         
         for (int i = 0; i < hdr.length; i++)
         {
            if (hdr[i].length() > maxlen)
               maxlen = hdr[i].length();
         }

         // create the frame definitions on the fly
         logosd = new ScreenDefinition(LOGO_FID);
         editsd = new ScreenDefinition(EDIT_FID);
   
         // logo frame configuration
         FrameConfig logofc = new FrameConfig(ComponentConfig.FRAME, null);
         logofc.setRow(0);
         logofc.setHeightChars(18);
         logofc.setWidthChars(maxlen);
         logofc.setBox(false);
         logofc.setCentered(true);
         logofc.setDcolor(Color.NORMAL);
         logofc.setSideLabels(true);
         logosd.addConfig(logofc, LOGO_FID);

         // logo text
         String fmt = "x(" + maxlen + ")"; 
         for (int i = 0; i < hdr.length; i++)
         {
            ControlTextConfig t01 = new ControlTextConfig(ComponentConfig.TEXT, null);
            t01.setDataType("character");
            t01.setFormat(fmt);
            logosd.addConfig(t01, LOGO_ID + i);
         }

         // edit frame configuration
         FrameConfig editfc = new FrameConfig(ComponentConfig.FRAME, null);
         editfc.setRow(19);
         editfc.setHeightChars(3);
         editfc.setWidthChars(38);
         editfc.setBox(false);
         editfc.setCentered(true);
         editfc.setOverlay(true);
         editfc.setDcolor(Color.NORMAL);
         editfc.setSideLabels(true);
         editsd.addConfig(editfc, EDIT_FID);
         
         FillInConfig fic;
         
         // user ID entry field
         fic = new FillInConfig(ComponentConfig.FILL_IN, null);
         fic.setDataType("character");
         fic.setFormat("x(25)");
         fic.setLabel("User Name");
         editsd.addConfig(fic, USER_ID);
         
         // set next widget position
         addSkip(editsd, 0);
   
         // password entry field
         fic = new FillInConfig(ComponentConfig.FILL_IN, null);
         fic.setDataType("character");
         fic.setFormat("x(25)");
         fic.setLabel(" Password");
         fic.setBlank(true);
         editsd.addConfig(fic, PASSWORD_ID);

         // instantiate screen data
         client.pushScreenDefinition(new ScreenDefinition[] {logosd});
         client.pushScreenDefinition(new ScreenDefinition[] {editsd});

         // set status line for input operation
         client.statusInput(STD_MSG);
      }
      else
      {
         String msgtxt = INVALID_PASSWORD;
         
         if (code == AUTH_RESULT_INVALID_USERID)
         {
            msgtxt = INVALID_USER;
         }
         
         // display error
         client.message(msgtxt);
      }

      // fill the screen buffer with the splash screen and info text
      for (int i = 0; i < hdr.length; i++)
         logosb.putWidgetValue(LOGO_ID + i, new character(hdr[i]));

      // process the edit frame screen buffer
      if (lastid != null)
      {
         // on errors, we have to display the previously typed userid
         editsb.putWidgetValue(USER_ID, new character(lastid));
      }
      editsb.putWidgetValue(INFO_ID, new character(option));

      String id = null;
      String pw = null;
                                                   
      EventList el = new EventList(true);
      ScreenBuffer sbf = null;

      while (true)
      {                                             
         // drive the logon frame I/O
         tk.setInvalidate(true);
         client.enable(EDIT_FID, null, editsb, true);
         client.clear(EDIT_FID, true);                         
         client.view(LOGO_FID, logosb, null, true);
         client.putScreen(option, 23, 7, Color.MESSAGES);
         client.view(EDIT_FID, editsb, null, true);
         editsb.resetChanged();
         tk.setInvalidate(false);

         // refresh insert indicator (workaround)
         if (Window.getInsertMode() == Window.INSERT_MODE_INS)
         {
            // turn insert mode off
            Window.setInsertMode(Window.INSERT_MODE_OFF, true);

            // apply insert key to user name fill-in, turning insert mode
            // on by this
            KeyInput e = new KeyInput(client.getWidget(USER_ID), 
                                      Key.VK_INSERT,
                                      Keyboard.KA_INSERT_MODE);
            WindowManager.topWindow().processEvent(e);
         }

         try
         {
            client.setAuthMode(true);
            sbf = client.waitFor(el, USER_ID, -1, editsb);
         }
         catch (ConditionException cx)
         {
            continue;
         }
         finally
         {
            client.setAuthMode(false);
         }

         tk.setInvalidate(true);
         client.enable(EDIT_FID, null, editsb, false);
         tk.setInvalidate(false);
   
         // inspect the input
         int keyCode = sbf.getKeyCode();

         if (Keyboard.keyFunction(keyCode).equals("END-ERROR"))
            continue;

         character idc = (character) sbf.getWidgetValue(USER_ID);
         
         if (idc == null && lastid == null)
            continue;

         id = (idc != null) ? idc.getValue().trim() : lastid;
         lastid = id;

         // quit?
         if (id.equals("quit"))
         {
            return new byte[] {};
         }

         character pwc = (character) sbf.getWidgetValue(PASSWORD_ID);
         
         if (pwc == null && lastpw == null)
         {
            pw = "";
         }
         else
         {
            pw = (pwc != null) ? pwc.getValue() : lastpw;
            lastpw = pw;
         }
         
         break; 
      }

      // return a byte array
      return SecurityManager.packageIdPassword(id, pw);      
   }
   
   /**
    * Finalizes any resources allocated during authentication by the client.
    */
   @Override
   public void clientFinalize()
   {
      client.statusInputRevert();
      client.hideAll(true);
      client.destroyFrame(LOGO_FID);
      client.destroyFrame(EDIT_FID);
   }

   /**
    * Not used for the client side.
    *
    * @param    auth
    *           The authorization input from the client.
    * @param    parameter
    *           Additional configuration parameters taken from the directory.
    *
    * @return   Always <code>null</code>.
    */
   @Override
   public AuthenticationResponse serverAuthHook(byte[] auth, String parameter)
   {
      return null;
   }

   /**
    * This has no effect as this class only run on client side.
    *
    * @return  Always <code>null</code>.
    */
   @Override
   public SessionListener getSessionListener()
   {
      return null;
   }

   /**
    * Configures the Authenticator by setting the "option" parameter from directory.xml.
    * This has no effect as this class only run on client side.
    *
    * @param   option
    *          The value of "option" entry for the auth plugin.
    */
   @Override
   public void configure(String option)
   {
   }

   /**
    * Returns a set of entities that this class handles. The list of entities / parameters will
    * be configured using other ways.
    * This has no effect as this class only run on client side.
    *
    * @return  always null
    */
   @Override
   public Set<String> getAuthenticationEntities()
   {
      return null;
   }

   /**
    * Add specified number of empty lines
    *
    * @param   sd
    *          The frame to which to add a skip.
    * @param   lines
    *          Lines to skip.
    */
   private void addSkip(ScreenDefinition sd, int lines)
   {
      SkipConfig sc = new SkipConfig(ComponentConfig.SKIP, null);
      sc.setVertical(true);   // skip not space
      sc.setHeight(lines);
      sd.addConfig(sc);
   }
}
